//
// Created by michal on 23.4.18.
//

#include <stdbool.h>
#include <math.h>
#include <soc/spi_reg.h>
#include <esp_clk.h>
#include <freertos/FreeRTOS.h>
#include <freertos/portmacro.h>

#include "spi_master.h"


#define BYTES_PER_DESCRIPTOR    (4096 - 4)

/**  CS pre-transmission delay
 * Unless specified in setup_time (in that case it sums up with setup_time)
 * Setting setup_time works only in half duplex mode due to hardware bug
 *
 * SPI_CLOCK_MODE_0: half a cycle before clock goes high
 * SPI_CLOCK_MODE_1: at the same time as clock goes high
 * SPI_CLOCK_MODE_2: half a cycle before clock goes low
 * SPI_CLOCK_MODE_3: at the same time as clock goes low
 */

typedef struct
{
    bool wr_bit_order;
    bool rd_bit_order;
    uint8_t mosi_delay_num;
    uint8_t mosi_delay_mode;
    uint8_t miso_delay_num;
    uint8_t miso_delay_mode;
    // bool cs_hold; TODO implement?
    // uint8_t hold_time; TODO implement?
    // uint8_t cs_hold_delay; TODO implement?
    bool clk_equ_sysclk;
    uint8_t clkdiv_pre;
    uint8_t clkcnt_n;
    uint8_t clkcnt_h;
    uint8_t clkcnt_l;
    bool ck_out_edge;
    //bool cs_keep_active; TODO implement?
    bool ck_idle_edge;
    bool usr_dummy_idle;
    // bool ck_dis; TODO implement?
    // uint8_t master_ck_sel; TODO implement?
} SPI_Master_Saved_Config_t;

typedef struct
{
    volatile uint32_t lock;
    SPI_Master_Saved_Config_t devices[3];
    intr_handle_t interrupt_handle;
} SPI_Master_Handle_t;

typedef struct descriptor
{
    uint32_t header;
    uint8_t * data_buffer;
    struct descriptor * next_descriptor;
} SPI_Master_DMA_Descriptor_t;

DMA_ATTR static SPI_Master_DMA_Descriptor_t read_dma_descriptor;
DMA_ATTR static SPI_Master_DMA_Descriptor_t write_dma_descriptor;

/* ------------------------------------------------------------------------------------------------------------------ */

static SPI_Master_Handle_t spi_master_handles[3] =
{
    {
        .lock = 0
    },
    {
        .lock = 0
    },
    {
        .lock = 0
    }
};

IRAM_ATTR static void SPI_Master_IRQ_DMA(void * arg)
{
    SPI_Device_t device = (SPI_Device_t)arg;

    // disable interrupts
    REG_SET_FIELD(SPI_SLAVE_REG(device), SPI_INT_EN, 0);

    // clear trans_done interrupt
    REG_CLR_BIT(SPI_SLAVE_REG(device), SPI_TRANS_DONE);

    free(write_dma_descriptor.next_descriptor);
    free(read_dma_descriptor.next_descriptor);

    SPI_DMA_Channel_Free(device);

    SPI_Master_Handle_t * target_handle = &spi_master_handles[device - 1];

    ESP_ERROR_CHECK(esp_intr_free(target_handle->interrupt_handle));

    target_handle->lock = 0;
}

// will be called when switching from one CS line to another to preserve custom config for every CS line
IRAM_ATTR static void SPI_Master_Load_Config(SPI_Device_t device, SPI_Master_CS_t cs_line)
{
    // should already be in lock here

    SPI_Master_Handle_t * target_handle = &spi_master_handles[device - 1];
    const SPI_Master_Saved_Config_t * target_config = &(target_handle->devices[cs_line]);

    if (target_config->wr_bit_order)
        REG_SET_BIT(SPI_CTRL_REG(device), SPI_WR_BIT_ORDER);
    else
        REG_CLR_BIT(SPI_CTRL_REG(device), SPI_WR_BIT_ORDER);

    if (target_config->rd_bit_order)
        REG_SET_BIT(SPI_CTRL_REG(device), SPI_RD_BIT_ORDER);
    else
        REG_CLR_BIT(SPI_CTRL_REG(device), SPI_RD_BIT_ORDER);

    REG_SET_FIELD(SPI_CTRL2_REG(device), SPI_MOSI_DELAY_NUM, target_config->mosi_delay_num);
    REG_SET_FIELD(SPI_CTRL2_REG(device), SPI_MOSI_DELAY_MODE, target_config->mosi_delay_mode);
    REG_SET_FIELD(SPI_CTRL2_REG(device), SPI_MISO_DELAY_NUM, target_config->miso_delay_num);
    REG_SET_FIELD(SPI_CTRL2_REG(device), SPI_MISO_DELAY_MODE, target_config->miso_delay_mode);

    if (target_config->clk_equ_sysclk)
        REG_SET_BIT(SPI_CLOCK_REG(device), SPI_CLK_EQU_SYSCLK);
    else
        REG_CLR_BIT(SPI_CLOCK_REG(device), SPI_CLK_EQU_SYSCLK);

    REG_SET_FIELD(SPI_CLOCK_REG(device), SPI_CLKDIV_PRE, target_config->clkdiv_pre);
    REG_SET_FIELD(SPI_CLOCK_REG(device), SPI_CLKCNT_N, target_config->clkcnt_n);
    REG_SET_FIELD(SPI_CLOCK_REG(device), SPI_CLKCNT_H, target_config->clkcnt_h);
    REG_SET_FIELD(SPI_CLOCK_REG(device), SPI_CLKCNT_L, target_config->clkcnt_l);

    if (target_config->ck_out_edge)
        REG_SET_BIT(SPI_USER_REG(device), SPI_CK_OUT_EDGE);
    else
        REG_CLR_BIT(SPI_USER_REG(device), SPI_CK_OUT_EDGE);

    if (target_config->ck_idle_edge)
        REG_SET_BIT(SPI_PIN_REG(device), SPI_CK_IDLE_EDGE);
    else
        REG_CLR_BIT(SPI_PIN_REG(device), SPI_CK_IDLE_EDGE);

/*
    if (target_config->cs_keep_active)
        REG_SET_BIT(SPI_PIN_REG(device), SPI_CS_KEEP_ACTIVE);
    else
        REG_CLR_BIT(SPI_PIN_REG(device), SPI_CS_KEEP_ACTIVE);
*/

    if (target_config->usr_dummy_idle)
        REG_SET_BIT(SPI_USER_REG(device), SPI_USR_DUMMY_IDLE);
    else
        REG_CLR_BIT(SPI_USER_REG(device), SPI_USR_DUMMY_IDLE);

    REG_SET_BITS(SPI_PIN_REG(device), ~(1u << cs_line), 7);
}

/* ------------------------------------------------------------------------------------------------------------------ */

esp_err_t SPI_Master_Config(SPI_Device_t device, SPI_Master_CS_t cs_line, const SPI_Master_Config_t * config)
{
    assert(SPI_DEVICE_CHECK(device));
    assert(SPI_MASTER_CS_CHECK(cs_line));
    assert(SPI_CLOCK_MODE_CHECK(config->clock_mode));
    assert(SPI_BIT_ORDER_CHECK(config->read_bit_order));
    assert(SPI_BIT_ORDER_CHECK(config->write_bit_order));

    uint32_t apb_frequency = esp_clk_apb_freq();

    assert(config->frequency <= apb_frequency);

    SPI_Master_Handle_t * target_handle = &(spi_master_handles[device - 1]);

    uint32_t set = 1;
    uxPortCompareSet(&target_handle->lock, 0, &set);
    if (set)
        return ESP_FAIL;

    SPI_Master_Saved_Config_t * target_config = &(target_handle->devices[cs_line]);

    // is not per-cs line configuration TODO implement?
    //target_config->cs_keep_active = config->flags & SPI_MASTER_FLAG_CS_KEEP_ENABLED;
    target_config->usr_dummy_idle = config->flags & SPI_MASTER_FLAG_DUMMY_CLOCK_DISABLE;

    if (config->flags & SPI_MASTER_FLAG_CS_ACTIVE_HIGH)
    {
        uint8_t tmp = REG_GET_FIELD(SPI_PIN_REG(device), SPI_MASTER_CS_POL);
        tmp |= 1u << cs_line;
        REG_SET_FIELD(SPI_PIN_REG(device), SPI_MASTER_CS_POL, tmp);
    } else
    {
        uint8_t tmp = REG_GET_FIELD(SPI_PIN_REG(device), SPI_MASTER_CS_POL);
        tmp &= ~(1u << cs_line);
        REG_SET_FIELD(SPI_PIN_REG(device), SPI_MASTER_CS_POL, tmp);
    }

    // set clock dividers
    unsigned target_clock_divider = (unsigned) roundf((float) apb_frequency / config->frequency);

    // TODO make duty cycle configurable?
    // setting too high value may cause SPI hardware to stop working

    if (target_clock_divider == 1)
    {
        target_config->clk_equ_sysclk = 1;
        target_config->clkdiv_pre = 0;
        target_config->clkcnt_n = 0;
        target_config->clkcnt_h = 0;
        target_config->clkcnt_l = 0;
    } else if (target_clock_divider == 2)
    {
        target_config->clk_equ_sysclk = 0;
        target_config->clkdiv_pre = 0;
        target_config->clkcnt_n = target_clock_divider - 1;
        target_config->clkcnt_h = 0;
        target_config->clkcnt_l = target_config->clkcnt_n;
    } else if (target_clock_divider <= 64)
    {
        target_config->clk_equ_sysclk = 0;
        target_config->clkdiv_pre = 0;
        target_config->clkcnt_n = target_clock_divider - 1;
        target_config->clkcnt_h = target_config->clkcnt_n / 2 - 1;
        target_config->clkcnt_l = target_config->clkcnt_n;
    } else
    {
        for (uint8_t i = 64; i > 0; i--)
            if (target_clock_divider % i == 0)
            {
                target_config->clk_equ_sysclk = 0;
                target_config->clkdiv_pre = target_clock_divider / i - 1;
                target_config->clkcnt_n = i - 1;

                // TODO min 20 %
                // max 98 %
                target_config->clkcnt_h = target_config->clkcnt_n / 2 - 1;
                target_config->clkcnt_l = target_config->clkcnt_n;
                break;
            }
    }

    target_config->miso_delay_mode = 0;
    target_config->miso_delay_num = 0;

    target_config->mosi_delay_mode = 0;
    target_config->mosi_delay_num = 0;

    switch (config->clock_mode)
    {
        case SPI_CLOCK_MODE_0:
            target_config->ck_idle_edge = 0;
            target_config->ck_out_edge = 0;
            break;
        case SPI_CLOCK_MODE_1:
            target_config->ck_idle_edge = 0;
            target_config->ck_out_edge = 1;
            break;
        case SPI_CLOCK_MODE_2:
            target_config->ck_idle_edge = 1;
            target_config->ck_out_edge = 1;
            break;
        case SPI_CLOCK_MODE_3:
            target_config->ck_idle_edge = 1;
            target_config->ck_out_edge = 0;
            break;
    }

    switch (config->write_bit_order)
    {
        case SPI_BIT_ORDER_LSB_FIRST:
            target_config->wr_bit_order = 1;
            break;
        case SPI_BIT_ORDER_MSB_FIRST:
            target_config->wr_bit_order = 0;
            break;
    }

    switch (config->read_bit_order)
    {
        case SPI_BIT_ORDER_LSB_FIRST:
            target_config->rd_bit_order = 1;
            break;
        case SPI_BIT_ORDER_MSB_FIRST:
            target_config->rd_bit_order = 0;
            break;
    }

    target_handle->lock = 0;

    return ESP_OK;
}

esp_err_t SPI_Master_Transmit(SPI_Device_t device,
                              SPI_Master_CS_t cs_line,
                              const SPI_Master_Transmission_t * transmission,
                              uint8_t * mosi_buffer,
                              uint8_t * miso_buffer)
{
    assert(SPI_DEVICE_CHECK(device));
    assert(SPI_MASTER_CS_CHECK(cs_line));
    assert(SPI_MASTER_MODE_CHECK(transmission->mode));
    assert((transmission->mosi_bits <= 8 * 32 && transmission->miso_bits <= 8 * 32) ||
           (transmission->mosi_bits == 0 && transmission->miso_bits <= 16 * 32)     ||
           (transmission->mosi_bits <= 16 * 32 && transmission->miso_bits == 0));
    assert(transmission->mosi_bits == 0 || transmission->mosi_bits >= transmission->miso_bits);
    assert(transmission->mosi_bits == 0 || mosi_buffer != NULL);
    assert(transmission->miso_bits == 0 || miso_buffer != NULL);
    assert(transmission->command_bits <= 16);
    assert(transmission->address_bits <= 64);
    assert(transmission->dummy_bits <= 64);
    assert(transmission->cs_pre_delay <= 15);
    // due to hardware bug CS setup time must be set to 0 while using command or address phase in full-duplex
    assert(transmission->cs_pre_delay == 0 || (transmission->command_bits == 0 && transmission->address_bits == 0) || transmission->mode != SPI_MASTER_MODE_FULL_DUPLEX);
    assert(transmission->cs_post_delay <= 15);
    // due to hardware bug dummy phase does not work while using mosi phase
    assert(transmission->mosi_bits == 0 || transmission->dummy_bits == 0);

    SPI_Master_Handle_t * target_handle = &spi_master_handles[device - 1];

    uint32_t set = 1;
    uxPortCompareSet(&target_handle->lock, 0, &set);
    if (set)
        return ESP_FAIL;

    SPI_Master_Saved_Config_t * target_config = &(target_handle->devices[cs_line]);

    // load config for target slave
    SPI_Master_Load_Config(device, cs_line);

    // set master mode
    REG_CLR_BIT(SPI_SLAVE_REG(device), SPI_SLAVE_MODE);

    // TODO do not use DMA
    //DPORT_SET_PERI_REG_BITS(DPORT_SPI_DMA_CHAN_SEL_REG, 3, 0, device * 2 - 2);

    switch (transmission->mode)
    {
        case SPI_MASTER_MODE_HALF_DUPLEX:
            // disable fast read mode
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_QIO);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_DIO);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_QUAD);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_DUAL);

            // enable 3-wire half duplex mode
            REG_SET_BIT(SPI_USER_REG(device), SPI_SIO);

            // disable all fast write modes
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_DIO);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_DUAL);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_QIO);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_QUAD);

            // disable full duplex mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_DOUTDIN);
            break;
        case SPI_MASTER_MODE_FULL_DUPLEX:
            // disable fast read mode
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_QIO);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_DIO);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_QUAD);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_DUAL);

            // disable 3-wire half duplex mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_SIO);

            // disable all fast write modes
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_DIO);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_DUAL);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_QIO);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_QUAD);

            // enable full duplex mode
            REG_SET_BIT(SPI_USER_REG(device), SPI_DOUTDIN);
            break;
        case SPI_MASTER_MODE_DOUT:
            // enable DOUT read mode
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_QIO);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_DIO);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_QUAD);
            REG_SET_BIT(SPI_CTRL_REG(device), SPI_FREAD_DUAL);

            // disable 3-wire half duplex mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_SIO);

            // enable DOUT write mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_DIO);
            REG_SET_BIT(SPI_USER_REG(device), SPI_FWRITE_DUAL);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_QIO);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_QUAD);

            // disable full duplex mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_DOUTDIN);
            break;
        case SPI_MASTER_MODE_DIO:
            // enable DIO read mode
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_QIO);
            REG_SET_BIT(SPI_CTRL_REG(device), SPI_FREAD_DIO);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_QUAD);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_DUAL);

            // disable 3-wire half duplex mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_SIO);

            // enable DIO write mode
            REG_SET_BIT(SPI_USER_REG(device), SPI_FWRITE_DIO);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_DUAL);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_QIO);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_QUAD);

            // disable full duplex mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_DOUTDIN);
            break;
        case SPI_MASTER_MODE_QOUT:
            // enable QOUT read mode
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_QIO);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_DIO);
            REG_SET_BIT(SPI_CTRL_REG(device), SPI_FREAD_QUAD);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_DUAL);

            // disable 3-wire half duplex mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_SIO);

            // enable QOUT write mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_DIO);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_DUAL);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_QIO);
            REG_SET_BIT(SPI_USER_REG(device), SPI_FWRITE_QUAD);

            // disable full duplex mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_DOUTDIN);
            break;
        case SPI_MASTER_MODE_QIO:
            // enable QIO read mode
            REG_SET_BIT(SPI_CTRL_REG(device), SPI_FREAD_QIO);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_DIO);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_QUAD);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_DUAL);

            // disable 3-wire half duplex mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_SIO);

            // enable QIO write mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_DIO);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_DUAL);
            REG_SET_BIT(SPI_USER_REG(device), SPI_FWRITE_QIO);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_QUAD);

            // disable full duplex mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_DOUTDIN);
            break;
    }

    if (transmission->command_bits > 0)
    {
        if (target_config->wr_bit_order == SPI_BIT_ORDER_MSB_FIRST) {
            uint16_t tmp_command = transmission->command << (16 - transmission->command_bits);
            REG_SET_FIELD(SPI_USER2_REG(device), SPI_USR_COMMAND_VALUE, (tmp_command >> 8) | (tmp_command << 8));
        } else if (target_config->wr_bit_order == SPI_BIT_ORDER_LSB_FIRST)
            REG_SET_FIELD(SPI_USER2_REG(device), SPI_USR_COMMAND_VALUE, transmission->command);

        REG_SET_FIELD(SPI_USER2_REG(device), SPI_USR_COMMAND_BITLEN, transmission->command_bits - 1);
        REG_SET_BIT(SPI_USER_REG(device), SPI_USR_COMMAND);
    } else
        REG_CLR_BIT(SPI_USER_REG(device), SPI_USR_COMMAND);

    if (transmission->address_bits > 0)
    {
        REG_SET_BIT(SPI_USER_REG(device), SPI_USR_ADDR);
        REG_SET_FIELD(SPI_USER1_REG(device), SPI_USR_ADDR_BITLEN, transmission->address_bits - 1);

        if (target_config->wr_bit_order == SPI_BIT_ORDER_MSB_FIRST)
        {
            if (transmission->address_bits > 32)
            {
                REG_WRITE(SPI_ADDR_REG(device), transmission->address >> (transmission->address_bits - 32));
                REG_WRITE(SPI_SLV_WR_STATUS_REG(device), transmission->address << (64 - transmission->address_bits));
            } else
                REG_WRITE(SPI_ADDR_REG(device), transmission->address << (32 - transmission->address_bits));
        } else if (target_config->wr_bit_order == SPI_BIT_ORDER_LSB_FIRST)
        {
            union
            {
                uint32_t value;
                uint8_t bytes[4];
            } byte_converter;

            byte_converter.value = transmission->address;
            uint8_t tmp = byte_converter.bytes[0];
            byte_converter.bytes[0] = byte_converter.bytes[3];
            byte_converter.bytes[3] = tmp;

            tmp = byte_converter.bytes[1];
            byte_converter.bytes[1] = byte_converter.bytes[2];
            byte_converter.bytes[2] = tmp;

            REG_WRITE(SPI_ADDR_REG(device), byte_converter.value);

            byte_converter.value = transmission->address >> 32u;
            tmp = byte_converter.bytes[0];
            byte_converter.bytes[0] = byte_converter.bytes[3];
            byte_converter.bytes[3] = tmp;

            tmp = byte_converter.bytes[1];
            byte_converter.bytes[1] = byte_converter.bytes[2];
            byte_converter.bytes[2] = tmp;

            REG_WRITE(SPI_SLV_WR_STATUS_REG(device), byte_converter.value);
        }
    } else
        REG_CLR_BIT(SPI_USER_REG(device), SPI_USR_ADDR);

    if (transmission->dummy_bits > 0)
    {
        REG_SET_BIT(SPI_USER_REG(device), SPI_USR_DUMMY);
        REG_SET_FIELD(SPI_USER1_REG(device), SPI_USR_DUMMY_CYCLELEN, transmission->dummy_bits - 1);
    } else
        REG_CLR_BIT(SPI_USER_REG(device), SPI_USR_DUMMY);

    if (transmission->miso_bits > 0)
    {
        REG_SET_BIT(SPI_USER_REG(device), SPI_USR_MISO);
        REG_CLR_BIT(SPI_USER_REG(device), SPI_USR_MISO_HIGHPART);

        REG_WRITE(SPI_MISO_DLEN_REG(device), transmission->miso_bits - 1);
    } else
        REG_CLR_BIT(SPI_USER_REG(device), SPI_USR_MISO);

    if (transmission->mosi_bits > 0)
    {
        if (transmission->mosi_bits > 8 * 32)
            // in this case data cannot be read
            REG_CLR_BIT(SPI_USER_REG(device), SPI_USR_MOSI_HIGHPART);
        else
            REG_SET_BIT(SPI_USER_REG(device), SPI_USR_MOSI_HIGHPART);

        REG_WRITE(SPI_MOSI_DLEN_REG(device), transmission->mosi_bits - 1);
        REG_SET_BIT(SPI_USER_REG(device), SPI_USR_MOSI);

    } else
        REG_CLR_BIT(SPI_USER_REG(device), SPI_USR_MOSI);

    if (transmission->cs_pre_delay > 0)
    {
        REG_SET_FIELD(SPI_CTRL2_REG(device), SPI_SETUP_TIME, transmission->cs_pre_delay);
        REG_SET_BIT(SPI_USER_REG(device), SPI_CS_SETUP);
    } else
        REG_CLR_BIT(SPI_USER_REG(device), SPI_CS_SETUP);

    REG_SET_FIELD(SPI_CTRL2_REG(device), SPI_HOLD_TIME, transmission->cs_post_delay);
    REG_SET_BIT(SPI_USER_REG(device), SPI_CS_HOLD);

    volatile uint32_t * fifo = (volatile uint32_t *)
            (transmission->mosi_bits > 8 * 32 ? SPI_W0_REG(device) : SPI_W8_REG(device));
    uint8_t buffer_pos = 0;
    uint32_t copied_bits = 0;

    while (copied_bits < transmission->mosi_bits)
    {
        uint32_t tmp = 0;

        for (uint8_t i = 0; i < 4; i++)
        {
            if (copied_bits >= transmission->mosi_bits)
                break;
            tmp |= mosi_buffer[buffer_pos++] << 8 * i;

            copied_bits += 8;
        }
        *fifo++ = tmp;
    }

    // disable all interrupts
    REG_SET_FIELD(SPI_SLAVE_REG(device), SPI_INT_EN, 0);

    // start user command
    REG_SET_BIT(SPI_CMD_REG(device), SPI_USR);

    // wait for SPI state machine to perform transmission
    while (REG_GET_FIELD(SPI_EXT2_REG(device), SPI_ST));

    copied_bits = 0;
    fifo = (volatile uint32_t *)SPI_W0_REG(device);
    buffer_pos = 0;

    while (copied_bits < transmission->miso_bits)
    {
        for (uint8_t i = 0; i < 4; i++)
        {
            if (copied_bits >= transmission->miso_bits)
                break;

            // when miso_bits is not multiple of 8 hardware will clear remaining bits itself
            miso_buffer[buffer_pos++] = (*fifo >> 8 * i) & 0xFF;

            copied_bits += 8;
        }
        fifo++;
    }

    target_handle->lock = 0;

    return ESP_OK;
}

// TODO is there any way to check word alignment of mosi & miso buffer?
esp_err_t SPI_Master_Transmit_DMA(SPI_Device_t device,
                                  SPI_Master_CS_t cs_line,
                                  const SPI_Master_Transmission_t * transmission,
                                  uint8_t * mosi_buffer,
                                  uint8_t * miso_buffer)
{
    assert(SPI_DEVICE_CHECK(device));
    assert(SPI_MASTER_CS_CHECK(cs_line));
    assert(SPI_MASTER_MODE_CHECK(transmission->mode));
    // due to hardware bug it is not possible to use half-duplex with both read and write phases while using DMA
    // TODO check this
    assert(transmission->mode == SPI_MASTER_MODE_FULL_DUPLEX || transmission->mosi_bits == 0 || transmission->miso_bits == 0);
    assert(transmission->mosi_bits == 0 || transmission->mosi_bits >= transmission->miso_bits);
    assert(transmission->mosi_bits == 0 || mosi_buffer != NULL);
    assert(transmission->miso_bits == 0 || miso_buffer != NULL);
    assert(transmission->command_bits <= 16);
    assert(transmission->address_bits <= 64);
    assert(transmission->dummy_bits <= 64);
    assert(transmission->mosi_bits <= 16777216); // fits in 24 bits
    assert(transmission->miso_bits <= 16777216);
    assert(transmission->cs_pre_delay <= 15);
    // due to hardware bug CS setup time must be set to 0 while using command or address phase
    assert(transmission->cs_pre_delay == 0 || (transmission->command_bits == 0 && transmission->address_bits == 0));
    assert(transmission->cs_post_delay <= 15);
    // due to hardware bug dummy phase does not work while using mosi phase
    assert(transmission->mosi_bits == 0 || transmission->dummy_bits == 0);
    assert(mosi_buffer == NULL || (esp_ptr_dma_capable(mosi_buffer) && esp_ptr_dma_capable(mosi_buffer + (transmission->mosi_bits + 7) / 8 - 1)));
    assert(miso_buffer == NULL || (esp_ptr_dma_capable(miso_buffer) && esp_ptr_dma_capable(miso_buffer + (transmission->miso_bits + 7) / 8 - 1)));
    // TODO mosi & miso buffers do not overlap

    SPI_Master_Handle_t * target_handle = &spi_master_handles[device - 1];

    uint32_t set = 1;
    uxPortCompareSet(&target_handle->lock, 0, &set);
    if (set)
        return ESP_FAIL;

    // try to acquire one of two SPI DMA channels
    if (SPI_DMA_Channel_Try_Acquire(device) != ESP_OK)
    {
        target_handle->lock = 0;
        return ESP_FAIL;
    }

    if (esp_intr_alloc(ETS_SPI0_INTR_SOURCE + device,
        ESP_INTR_FLAG_LOWMED | ESP_INTR_FLAG_IRAM, SPI_Master_IRQ_DMA,
        (void *)device,
        &target_handle->interrupt_handle) != ESP_OK)
    {
        SPI_DMA_Channel_Free(device);
        target_handle->lock = 0;
        return ESP_FAIL;
    }

    SPI_Master_Saved_Config_t * target_config = &(target_handle->devices[cs_line]);

    // reset DMA
    REG_SET_BIT(SPI_DMA_CONF_REG(device),
        SPI_IN_RST | SPI_OUT_RST | SPI_AHBM_FIFO_RST | SPI_AHBM_RST);

    REG_CLR_BIT(SPI_DMA_OUT_LINK_REG(device), SPI_OUTLINK_START);
    REG_CLR_BIT(SPI_DMA_IN_LINK_REG(device), SPI_INLINK_START);

    REG_CLR_BIT(SPI_DMA_CONF_REG(device),
        SPI_IN_RST | SPI_OUT_RST | SPI_AHBM_FIFO_RST | SPI_AHBM_RST);

    REG_SET_BIT(SPI_DMA_CONF_REG(device), SPI_OUT_DATA_BURST_EN);

    // fill DMA descriptors
    if (transmission->mosi_bits > 0)
    {
        SPI_Master_DMA_Descriptor_t * current_descriptor = &write_dma_descriptor;
        SPI_Master_DMA_Descriptor_t * descriptors = NULL;
        uint8_t * buffer = mosi_buffer;
        uint32_t remaining_bits = transmission->mosi_bits;

        if (remaining_bits > BYTES_PER_DESCRIPTOR * 8)
            descriptors = heap_caps_malloc(
                sizeof(SPI_Master_DMA_Descriptor_t) * (((transmission->mosi_bits + 7) / 8 + BYTES_PER_DESCRIPTOR - 1) / BYTES_PER_DESCRIPTOR - 1),
                MALLOC_CAP_DMA);

        while (remaining_bits > BYTES_PER_DESCRIPTOR * 8)
        {
            // owner = DMA controller = 1
            // eof = 0
            // size = length = 4096 - 4
            current_descriptor->header = 0x80FFCFFC;
            current_descriptor->data_buffer = buffer;
            current_descriptor->next_descriptor = descriptors++;

            current_descriptor = current_descriptor->next_descriptor;
            remaining_bits -= BYTES_PER_DESCRIPTOR * 8;
            buffer += BYTES_PER_DESCRIPTOR;
        }

        uint16_t bytes = ((remaining_bits + 7) / 8) & 0xFFF;
        uint16_t aligned_bytes = ((bytes + 3) & (~3)) & 0xFFF;

        current_descriptor->header = 0xC0000000 | (bytes << 12) | aligned_bytes;
        current_descriptor->data_buffer = buffer;
        current_descriptor->next_descriptor = NULL;

        REG_SET_FIELD(SPI_DMA_OUT_LINK_REG(device), SPI_OUTLINK_ADDR, (uint32_t)&write_dma_descriptor);
        REG_SET_BIT(SPI_DMA_OUT_LINK_REG(device), SPI_OUTLINK_START);
    }

    if (transmission->miso_bits > 0)
    {
        SPI_Master_DMA_Descriptor_t * current_descriptor = &read_dma_descriptor;
        SPI_Master_DMA_Descriptor_t * descriptors = NULL;
        uint8_t * buffer = miso_buffer;
        uint32_t remaining_bits = transmission->miso_bits;

        if (remaining_bits > BYTES_PER_DESCRIPTOR * 8)
            descriptors = heap_caps_malloc(
                sizeof(SPI_Master_DMA_Descriptor_t) * (((transmission->miso_bits + 7) / 8 + BYTES_PER_DESCRIPTOR - 1) / BYTES_PER_DESCRIPTOR - 1),
                MALLOC_CAP_DMA);

        while (remaining_bits > BYTES_PER_DESCRIPTOR * 8)
        {
            // owner = DMA controller = 1
            // eof = 0
            // size = length = 4096 - 4
            current_descriptor->header = 0x80FFCFFC;
            current_descriptor->data_buffer = buffer;
            current_descriptor->next_descriptor = descriptors++;

            current_descriptor = current_descriptor->next_descriptor;
            remaining_bits -= BYTES_PER_DESCRIPTOR * 8;
            buffer += BYTES_PER_DESCRIPTOR;
        }

        uint16_t bytes = ((remaining_bits + 7) / 8) & 0xFFF;
        uint16_t aligned_bytes = ((bytes + 3) & (~3)) & 0xFFF;

        current_descriptor->header = 0xC0000000 | (bytes << 12) | aligned_bytes;
        current_descriptor->data_buffer = buffer;
        current_descriptor->next_descriptor = NULL;

        REG_SET_FIELD(SPI_DMA_IN_LINK_REG(device), SPI_INLINK_ADDR, (uint32_t)&read_dma_descriptor);
        REG_SET_BIT(SPI_DMA_IN_LINK_REG(device), SPI_INLINK_START);
    } else
    {
        // workaround for SPI DMA hardware bug
        REG_SET_FIELD(SPI_DMA_IN_LINK_REG(device), SPI_INLINK_ADDR, 0);
        REG_SET_BIT(SPI_DMA_IN_LINK_REG(device), SPI_INLINK_START);
    }

    // load config for target slave
    SPI_Master_Load_Config(device, cs_line);

    // set master mode
    REG_CLR_BIT(SPI_SLAVE_REG(device), SPI_SLAVE_MODE);

    switch (transmission->mode)
    {
        case SPI_MASTER_MODE_HALF_DUPLEX:
            // disable fast read mode
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_QIO);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_DIO);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_QUAD);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_DUAL);

            // enable 3-wire half duplex mode
            REG_SET_BIT(SPI_USER_REG(device), SPI_SIO);

            // disable all fast write modes
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_DIO);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_DUAL);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_QIO);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_QUAD);

            // disable full duplex mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_DOUTDIN);
            break;
        case SPI_MASTER_MODE_FULL_DUPLEX:
            // disable fast read mode
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_QIO);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_DIO);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_QUAD);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_DUAL);

            // disable 3-wire half duplex mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_SIO);

            // disable all fast write modes
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_DIO);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_DUAL);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_QIO);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_QUAD);

            // enable full duplex mode
            REG_SET_BIT(SPI_USER_REG(device), SPI_DOUTDIN);
            break;
        case SPI_MASTER_MODE_DOUT:
            // enable DOUT read mode
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_QIO);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_DIO);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_QUAD);
            REG_SET_BIT(SPI_CTRL_REG(device), SPI_FREAD_DUAL);

            // disable 3-wire half duplex mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_SIO);

            // enable DOUT write mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_DIO);
            REG_SET_BIT(SPI_USER_REG(device), SPI_FWRITE_DUAL);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_QIO);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_QUAD);

            // disable full duplex mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_DOUTDIN);
            break;
        case SPI_MASTER_MODE_DIO:
            // enable DIO read mode
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_QIO);
            REG_SET_BIT(SPI_CTRL_REG(device), SPI_FREAD_DIO);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_QUAD);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_DUAL);

            // disable 3-wire half duplex mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_SIO);

            // enable DIO write mode
            REG_SET_BIT(SPI_USER_REG(device), SPI_FWRITE_DIO);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_DUAL);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_QIO);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_QUAD);

            // disable full duplex mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_DOUTDIN);
            break;
        case SPI_MASTER_MODE_QOUT:
            // enable QOUT read mode
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_QIO);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_DIO);
            REG_SET_BIT(SPI_CTRL_REG(device), SPI_FREAD_QUAD);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_DUAL);

            // disable 3-wire half duplex mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_SIO);

            // enable QOUT write mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_DIO);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_DUAL);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_QIO);
            REG_SET_BIT(SPI_USER_REG(device), SPI_FWRITE_QUAD);

            // disable full duplex mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_DOUTDIN);
            break;
        case SPI_MASTER_MODE_QIO:
            // enable QIO read mode
            REG_SET_BIT(SPI_CTRL_REG(device), SPI_FREAD_QIO);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_DIO);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_QUAD);
            REG_CLR_BIT(SPI_CTRL_REG(device), SPI_FREAD_DUAL);

            // disable 3-wire half duplex mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_SIO);

            // enable QIO write mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_DIO);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_DUAL);
            REG_SET_BIT(SPI_USER_REG(device), SPI_FWRITE_QIO);
            REG_CLR_BIT(SPI_USER_REG(device), SPI_FWRITE_QUAD);

            // disable full duplex mode
            REG_CLR_BIT(SPI_USER_REG(device), SPI_DOUTDIN);
            break;
    }

    if (transmission->command_bits > 0)
    {
        if (target_config->wr_bit_order == SPI_BIT_ORDER_MSB_FIRST) {
            uint16_t tmp_command = transmission->command << (16 - transmission->command_bits);
            REG_SET_FIELD(SPI_USER2_REG(device), SPI_USR_COMMAND_VALUE, (tmp_command >> 8) | (tmp_command << 8));
        } else if (target_config->wr_bit_order == SPI_BIT_ORDER_LSB_FIRST)
            REG_SET_FIELD(SPI_USER2_REG(device), SPI_USR_COMMAND_VALUE, transmission->command);

        REG_SET_FIELD(SPI_USER2_REG(device), SPI_USR_COMMAND_BITLEN, transmission->command_bits - 1);
        REG_SET_BIT(SPI_USER_REG(device), SPI_USR_COMMAND);
    } else
        REG_CLR_BIT(SPI_USER_REG(device), SPI_USR_COMMAND);

    if (transmission->address_bits > 0)
    {
        REG_SET_BIT(SPI_USER_REG(device), SPI_USR_ADDR);
        REG_SET_FIELD(SPI_USER1_REG(device), SPI_USR_ADDR_BITLEN, transmission->address_bits - 1);

        if (target_config->wr_bit_order == SPI_BIT_ORDER_MSB_FIRST)
        {
            if (transmission->address_bits > 32)
            {
                REG_WRITE(SPI_ADDR_REG(device), transmission->address >> (transmission->address_bits - 32));
                REG_WRITE(SPI_SLV_WR_STATUS_REG(device), transmission->address << (64 - transmission->address_bits));
            } else
                REG_WRITE(SPI_ADDR_REG(device), transmission->address << (32 - transmission->address_bits));
        } else if (target_config->wr_bit_order == SPI_BIT_ORDER_LSB_FIRST)
        {
            union
            {
                uint32_t value;
                uint8_t bytes[4];
            } byte_converter;

            byte_converter.value = transmission->address;
            uint8_t tmp = byte_converter.bytes[0];
            byte_converter.bytes[0] = byte_converter.bytes[3];
            byte_converter.bytes[3] = tmp;

            tmp = byte_converter.bytes[1];
            byte_converter.bytes[1] = byte_converter.bytes[2];
            byte_converter.bytes[2] = tmp;

            REG_WRITE(SPI_ADDR_REG(device), byte_converter.value);

            byte_converter.value = transmission->address >> 32u;
            tmp = byte_converter.bytes[0];
            byte_converter.bytes[0] = byte_converter.bytes[3];
            byte_converter.bytes[3] = tmp;

            tmp = byte_converter.bytes[1];
            byte_converter.bytes[1] = byte_converter.bytes[2];
            byte_converter.bytes[2] = tmp;

            REG_WRITE(SPI_SLV_WR_STATUS_REG(device), byte_converter.value);
        }
    } else
        REG_CLR_BIT(SPI_USER_REG(device), SPI_USR_ADDR);

    if (transmission->dummy_bits > 0)
    {
        REG_SET_BIT(SPI_USER_REG(device), SPI_USR_DUMMY);
        REG_SET_FIELD(SPI_USER1_REG(device), SPI_USR_DUMMY_CYCLELEN, transmission->dummy_bits - 1);
    } else
        REG_CLR_BIT(SPI_USER_REG(device), SPI_USR_DUMMY);

    if (transmission->mosi_bits > 0)
    {
        REG_SET_BIT(SPI_USER_REG(device), SPI_USR_MOSI);
        REG_WRITE(SPI_MOSI_DLEN_REG(device), transmission->mosi_bits - 1);
    } else
        REG_CLR_BIT(SPI_USER_REG(device), SPI_USR_MOSI);

    if (transmission->miso_bits > 0)
    {
        REG_SET_BIT(SPI_USER_REG(device), SPI_USR_MISO);
        REG_WRITE(SPI_MISO_DLEN_REG(device), transmission->miso_bits - 1);
    } else
        REG_CLR_BIT(SPI_USER_REG(device), SPI_USR_MISO);

    if (transmission->cs_pre_delay > 0)
    {
        REG_SET_FIELD(SPI_CTRL2_REG(device), SPI_SETUP_TIME, transmission->cs_pre_delay);
        REG_SET_BIT(SPI_USER_REG(device), SPI_CS_SETUP);
    } else
        REG_CLR_BIT(SPI_USER_REG(device), SPI_CS_SETUP);


    REG_SET_FIELD(SPI_CTRL2_REG(device), SPI_HOLD_TIME, transmission->cs_post_delay);
    REG_SET_BIT(SPI_USER_REG(device), SPI_CS_HOLD);

    // disable interrupts
    REG_SET_FIELD(SPI_SLAVE_REG(device), SPI_INT_EN, 0);

    // clear trans_done interrupt
    REG_CLR_BIT(SPI_SLAVE_REG(device), SPI_TRANS_DONE);

    // enable trans_done interrupt
    REG_SET_FIELD(SPI_SLAVE_REG(device), SPI_INT_EN, 0x10);

    // start user command
    REG_SET_BIT(SPI_CMD_REG(device), SPI_USR);

    return ESP_OK;
}

bool SPI_Master_Busy(SPI_Device_t device)
{
    assert(SPI_DEVICE_CHECK(device));

    return (spi_master_handles[device - 1].lock == 1);
}
