//
// Created by michal on 23.4.18.
//

#pragma once

#ifdef __cplusplus
extern "C" {
#endif

#include <esp_err.h>
#include <stdint.h>
#include <esp_attr.h>

#include "spi.h"

#define SPI_MASTER_MODE_CHECK(MODE)                (((MODE) == SPI_MASTER_MODE_HALF_DUPLEX) || \
                                                    ((MODE) == SPI_MASTER_MODE_FULL_DUPLEX) || \
                                                    ((MODE) == SPI_MASTER_MODE_DOUT)        || \
                                                    ((MODE) == SPI_MASTER_MODE_DIO)         || \
                                                    ((MODE) == SPI_MASTER_MODE_QOUT)        || \
                                                    ((MODE) == SPI_MASTER_MODE_QIO))

#define SPI_MASTER_CS_CHECK(CS)                    (((CS) == SPI_MASTER_CS0) || \
                                                    ((CS) == SPI_MASTER_CS1) || \
                                                    ((CS) == SPI_MASTER_CS2))

/**  example of half duplex transmission
 *            command                address                 write              read
 * MOSI   C3  C2  C1  C0  |  A5  A4  A3  A2  A1  A0  |  W3  W2  W1  W0  |  R3  R2  R1  R0
 */

/**  example of full duplex transmission
 *            command                address                     write & read
 * MOSI   C3  C2  C1  C0  |  A5  A4  A3  A2  A1  A0  |  W7  W6  W5  W4  W3  W2  W1  W0
 * MISO   --  --  --  --  |  --  --  --  --  --  --  |  R7  R6  R5  R4  R3  --  --  --
 */

/**  example of DOUT transmission
 *            command                    address                   write          read
 * MOSI   C3  C2  C1  C0  |  A7  A6  A5  A4  A3  A2  A1  A0  |  W4  W2  W0  |  R4  R2  R0
 * MISO   --  --  --  --  |  --  --  --  --  --  --  --  --  |  W5  W3  W1  |  R5  R3  R1
 */

/**  example of DIO transmission
 *            command            address           write          read
 * MOSI   C3  C2  C1  C0  |  A6  A4  A2  A0  |  W4  W2  W0  |  R4  R2  R0
 * MISO   --  --  --  --  |  A7  A5  A3  A1  |  W5  W3  W1  |  R5  R3  R1
 */

/**  example of QOUT transmission
 *            command                address             write      read
 * MOSI   C3  C2  C1  C0  |  A5  A4  A3  A2  A1  A0  |  W4  W0  |  R4  R0
 * MISO   --  --  --  --  |  --  --  --  --  --  --  |  W5  W1  |  R5  R1
 * WP     --  --  --  --  |  --  --  --  --  --  --  |  W6  W2  |  R6  R2
 * HOLD   --  --  --  --  |  --  --  --  --  --  --  |  W7  W3  |  R7  R3
 */

/**  example of QIO transmission
 *            command        address     write      read
 * MOSI   C3  C2  C1  C0  |  A4  A0  |  W4  W0  |  R4  R0
 * MISO   --  --  --  --  |  A5  A1  |  W5  W1  |  R5  R1
 * WP     --  --  --  --  |  A6  A2  |  W6  W2  |  R6  R2
 * HOLD   --  --  --  --  |  A7  A3  |  W7  W3  |  R7  R3
 */

typedef enum
{
    SPI_MASTER_MODE_HALF_DUPLEX,
    SPI_MASTER_MODE_FULL_DUPLEX,
    SPI_MASTER_MODE_DOUT,
    SPI_MASTER_MODE_DIO,
    SPI_MASTER_MODE_QOUT,
    SPI_MASTER_MODE_QIO
} SPI_Master_Mode_t;

typedef enum
{
    SPI_MASTER_CS0 = 0,
    SPI_MASTER_CS1,
    SPI_MASTER_CS2
} SPI_Master_CS_t;


/** keep (all!!) CS lines always active
#define SPI_MASTER_FLAG_CS_KEEP_ENABLED         0x0001u
 */

/** hold WP signal low while idle
 * DOES NOT WORK
#define SPI_MASTER_FLAG_WP_IDLE_LOW             0x0002u
 */

/** disable clock signal in dummy phase */
#define SPI_MASTER_FLAG_DUMMY_CLOCK_DISABLE     0x0004u

/** make CS signal active high */
#define SPI_MASTER_FLAG_CS_ACTIVE_HIGH          0x0008u


typedef struct
{
    SPI_Clock_Mode_t clock_mode;
    SPI_Bit_Order_t read_bit_order;
    SPI_Bit_Order_t write_bit_order;
    uint32_t frequency;
    uint32_t flags;
} SPI_Master_Config_t;

typedef struct
{
    SPI_Master_Mode_t mode;
    uint16_t command;
    uint8_t command_bits;   /**< 0 - 16 */
    uint64_t address;
    uint8_t address_bits;   /**< 0 - 64 */
    uint8_t dummy_bits;     /**< 0 - 64 */
    uint32_t mosi_bits;
    uint32_t miso_bits;
    uint8_t cs_pre_delay;   /**< 0 - 15 */
    uint8_t cs_post_delay;  /**< 0 - 15 */
} SPI_Master_Transmission_t;

/**
 *
 * @param device
 * @param cs_line
 * @param config
 * @return
 */
IRAM_ATTR esp_err_t SPI_Master_Config(SPI_Device_t device, SPI_Master_CS_t cs_line, const SPI_Master_Config_t * config);

/**
 *
 * @param device
 * @param cs_line
 * @param transmission
 * @param mosi_buffer
 * @param miso_buffer
 * @return
 */
IRAM_ATTR esp_err_t SPI_Master_Transmit(SPI_Device_t device,
                                        SPI_Master_CS_t cs_line,
                                        const SPI_Master_Transmission_t * transmission,
                                        uint8_t * mosi_buffer,
                                        uint8_t * miso_buffer);

IRAM_ATTR esp_err_t SPI_Master_Transmit_DMA(SPI_Device_t device,
                                            SPI_Master_CS_t cs_line,
                                            const SPI_Master_Transmission_t * transmission,
                                            uint8_t * mosi_buffer,
                                            uint8_t * miso_buffer);

IRAM_ATTR bool SPI_Master_Busy(SPI_Device_t device);

#ifdef __cplusplus
}
#endif
